from typing import Union, Type, overload, Tuple, Dict, Any


ParamRet = Tuple[Tuple[Union["Insertable", Type["Insertable"]]], Dict[str, Any]]


class TemplateEvent:
    """
    事件基类, 用于承载事件信息
    """

    @staticmethod
    def param_export(
            *args: Union["Insertable", Type["Insertable"]],
            **kwargs
    ):
        """该静态函数用于包装参数"""
        return args, kwargs

    @staticmethod
    @overload
    def get_params() -> ParamRet:
        ...

    @classmethod
    @overload
    def get_params(cls) -> ParamRet:
        ...

    def get_params(self) -> ParamRet:
        """外部方法,用于返回该事件实例的内部参数;该方法可以是 `classmethod`

        需要返回的内容应当是以下形式的:
         - *插入器, 一类特殊的事件, 可以为其他事件提供额外参数
         - 参数类型名 = 实际参数, 当事件只返回一个该类参数时, 请用该参数的类型名
         - 参数默认名 = 实际参数, 当事件返回多个该类参数时, 请用默认名区分各参数
        """
        return self.param_export(
        )

    def get(self, attr_name: str):
        return getattr(self, attr_name, None)


class StructuredEvent(TemplateEvent):
    """
    可结构化的事件类
    """
    def __init__(self, **kwargs):
        for k, v in kwargs:
            setattr(self, k, v)


class Insertable(TemplateEvent):
    """用于标记插入器的事件类, 非继承该类的插入器会无法使用"""
    pass



from typing import Any, overload, List, Optional
from ..utils import ArgumentPackage
from abc import ABC, abstractmethod


class TemplateDecorator(ABC):
    """
    参数装饰器基类，用于修饰函数或事件

    Args:
        keep: 是否保留未装饰的参数,若为False则会覆盖原有参数
    """

    __disable__: List[str] = ""

    keep: bool = True
    may_target_type: type

    def __init__(self, target_type: Optional[type] = None, keep: bool = True):
        """
            参数装饰器基类，用于修饰函数或事件

            Args:
                target_type: 可能的目标参数类型,若填入则装饰器只接受该类型的参数
                keep: 是否保留未装饰的参数,若为False则会覆盖原有参数
            """
        self.may_target_type = target_type
        self.keep = keep

    def supply_wrapper(self, args_name: str, args_value: Any):
        result = self.supply(ArgumentPackage(args_name, args_value.__class__, args_value))
        if result is None:
            return
        if not self.keep:
            return {args_name: result}
        return {result.__class__.__name__: result}

    @staticmethod
    @abstractmethod
    @overload
    def supply(target_argument: ArgumentPackage) -> Any:
        ...

    @abstractmethod
    def supply(self, target_argument: ArgumentPackage) -> Any:
        """在回调函数的参数被解析后调用, 用于修饰函数的参数

        Args:
            target_argument: 一个保存了参数名、参数注解与参数内容的类
        Returns:
            装饰后的值
        """


class EventDecorator(TemplateDecorator):
    __disable__ = 'after_parser'

    @abstractmethod
    def supply(self, target_argument: ArgumentPackage) -> Any:
        return target_argument.value


class ArgsDecorator(TemplateDecorator):
    __disable__ = 'before_parser'

    @abstractmethod
    def supply(self, target_argument: ArgumentPackage) -> Any:
        return target_argument.value



import inspect
from functools import lru_cache
from typing import Type, Union, Callable, Any

from .entities.condition import TemplateCondition
from .entities.event import TemplateEvent

Event_T = Union[Type[TemplateEvent], TemplateEvent]
Condition_T = Union[Type[TemplateCondition], TemplateCondition]


class ArgumentPackage:
    name: str
    annotation: Any
    value: Any

    __slots__ = ("name", "value", "annotation")

    def __init__(self, name, annotation, value):
        self.name = name
        self.annotation = annotation
        self.value = value

    def __repr__(self):
        return (
            "<ArgumentPackage name={0} annotation={1} value={2}".format(
                self.name, self.annotation, self.value
            )
        )


async def run_always_await(callable_target, *args, **kwargs):
    if iscoroutinefunction(callable_target):
        return await callable_target(*args, **kwargs)
    return callable_target(*args, **kwargs)


@lru_cache(4096)
def argument_analysis(callable_target: Callable):
    return [
        (
            name,
            param.annotation if param.annotation is not inspect.Signature.empty else None,
            param.default if param.default is not inspect.Signature.empty else None,
        )
        for name, param in inspect.signature(callable_target).parameters.items()
    ]


@lru_cache(4096)
def iscoroutinefunction(o):
    return inspect.iscoroutinefunction(o)


def event_class_generator(target=TemplateEvent):
    for i in target.__subclasses__():
        yield i
        if i.__subclasses__():
            yield from event_class_generator(i)


def search_event(name: str):
    for i in event_class_generator():
        if i.__name__ == name:
            return i



import asyncio
from datetime import datetime
from typing import List, Union, Type, Dict, Any
from .entities.delegate import EventDelegate
from .entities.event import TemplateEvent
from .entities.publisher import Publisher
from .entities.decorator import TemplateDecorator
from .entities.subscriber import Subscriber
from .utils import Condition_T, search_event, event_class_generator


class EventSystem:
    loop: asyncio.AbstractEventLoop
    publisher_list: List[Publisher]
    safety_interval: float
    last_run: datetime
    current_event: Union[TemplateEvent, Dict[str, Any]]

    def __init__(
            self,
            *,
            loop: asyncio.AbstractEventLoop = None,
            interval: float = 0.00
    ):
        self.publisher_list = []
        self.loop = loop or asyncio.get_event_loop()
        self.safety_interval = interval
        self.last_run = datetime.now()

    def event_spread(self, target: Union[TemplateEvent, Dict[str, Any]]):
        if (datetime.now() - self.last_run).total_seconds() >= self.safety_interval:
            self.current_event = target
            try:
                for pub in self.publisher_generator():
                    pub.on_event(self)
            except asyncio.CancelledError:
                return
        self.last_run = datetime.now()

    def publisher_generator(self):
        able_pubs = list(
            filter(
                lambda x: all([condition.judge(self.current_event) for condition in x.external_conditions]),
                self.publisher_list
            )
        )
        able_pubs.sort(key=lambda x: x.priority)
        return able_pubs

    def get_publisher(self, target: Union[Type[TemplateEvent], Condition_T]):
        p_list = []
        for publisher in self.publisher_list:
            if target in publisher:
                p_list.append(publisher)
        if len(p_list) > 0:
            return p_list
        return False

    def remove_publisher(self, target: Publisher):
        self.publisher_list.remove(target)

    def register(
            self,
            event: Union[str, Type[TemplateEvent]],
            *,
            priority: int = 16,
            conditions: List[Condition_T] = None,
            decorators: List[TemplateDecorator] = None,
    ):
        if isinstance(event, str):
            name = event
            event = search_event(event)
            if not event:
                raise Exception(name + " cannot found!")

        events = [event]
        events.extend(event_class_generator(event))
        conditions = conditions or []
        decorators = decorators or []

        def register_wrapper(exec_target):
            if not isinstance(exec_target, Subscriber):
                exec_target = Subscriber(
                    callable_target=exec_target,
                    decorators=decorators
                )
            for e in events:
                may_publishers = self.get_publisher(e)
                _event_handler = EventDelegate(event=e)
                _event_handler += exec_target
                if not may_publishers:
                    self.publisher_list.append(Publisher(priority, conditions, _event_handler))
                else:
                    for m_publisher in may_publishers:
                        if m_publisher.equal_conditions(conditions):
                            m_publisher += _event_handler
                            break
                    else:
                        self.publisher_list.append(Publisher(priority, conditions, _event_handler))

            return exec_target

        return register_wrapper



from typing import Callable, Optional, List
from .decorator import TemplateDecorator
from ..utils import argument_analysis


class SubscriberInterface:
    callable_target: Callable


class Subscriber(SubscriberInterface):
    def __init__(
            self,
            callable_target: Callable,
            *,
            subscriber_name: Optional[str] = None,
            decorators: Optional[List[TemplateDecorator]] = None
    ) -> None:
        self.callable_target = callable_target
        self.subscriber_name = subscriber_name or callable_target.__name__
        self.decorators = decorators
        self.params = argument_analysis(self.callable_target)

    def __call__(self, *args, **kwargs):
        return self.callable_target(*args, **kwargs)

    def __repr__(self):
        return f'<Subscriber,name={self.subscriber_name}>'

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if isinstance(other, Subscriber):
            return other.name == self.subscriber_name
        elif isinstance(other, str):
            return other == self.subscriber_name

    @property
    def name(self):
        return self.subscriber_name

    @name.setter
    def name(self, new_name):
        self.subscriber_name = new_name

    @staticmethod
    def set(
            subscriber_name: Optional[str] = None,
            priority: int = 16,
            decorators: List[TemplateDecorator] = None
    ):
        """该方法生成一个订阅器实例，该订阅器负责调控装饰的可执行函数

        若订阅器的注册事件是消息链相关的，可预先根据写入的命令相关参数进行判别是否执行该订阅器装饰的函数

        若不填如下参数则表示接受任意形式的命令

        Args:
            subscriber_name :装饰器名字，可选
            priority: 优先级，默认为最高
            decorators: 装饰器列表，可选
        """

        def wrapper(func):
            subscriber = Subscriber(
                func,
                subscriber_name=subscriber_name,
                decorators=decorators
            )
            return subscriber

        return wrapper

    subscriber_name: Optional[str]
    decorators: Optional[List[TemplateDecorator]]



from typing import Union, Dict, Optional, List, Any, TYPE_CHECKING
from .delegate import EventDelegate, Subscriber
from ..utils import Condition_T, Event_T
from .event import TemplateEvent
from .condition import TemplateCondition

if TYPE_CHECKING:
    from .. import EventSystem


class Publisher:
    external_conditions: List[Condition_T]
    internal_delegate: Dict[str, EventDelegate]
    priority: int

    def __init__(
            self,
            priority: int = 16,
            conditions: Optional[List[Condition_T]] = None,
            *delegates: Optional[EventDelegate]
    ):
        self.external_conditions = conditions or []
        self.internal_delegate = {}
        self.priority = priority
        if delegates:
            for delegate in delegates:
                self.__iadd__(delegate)

    def __iadd__(self, other: EventDelegate):
        _name = other.bind_event.__name__
        if _name in self.internal_delegate:
            self.internal_delegate[_name] += other.subscribers
        else:
            self.internal_delegate[_name] = other
        return self

    def __getitem__(self, item: Union[TemplateEvent, str]):
        if isinstance(item, TemplateEvent):
            return self.internal_delegate[item.__class__.__name__]
        return self.internal_delegate[item]

    def __contains__(self, item: Union[Event_T, EventDelegate, Subscriber, Condition_T]):
        if isinstance(item, TemplateCondition):
            for condition in self.external_conditions:
                if item == condition:
                    return True
            return False
        elif isinstance(item, Subscriber):
            for delegate in self.internal_delegate.values():
                if item in delegate.subscribers:
                    return True
            return False
        elif isinstance(item, EventDelegate):
            return item in self.internal_delegate.values()
        else:
            return item.__name__ in self.internal_delegate

    def equal_conditions(self, cl: List[Condition_T]):
        cl.sort(key=lambda x: id(x))
        self.external_conditions.sort(key=lambda x: id(x))
        return cl == self.external_conditions

    def on_event(self, event_system: "EventSystem"):
        target: Union[TemplateEvent, Dict[str, Any]] = event_system.current_event
        event_name = target.__class__.__name__
        if isinstance(target, dict):
            event_name = target.get('type')
        if event_name in self.internal_delegate.keys():
            if isinstance(target, TemplateEvent):
                event_system.loop.create_task(self.internal_delegate[event_name].executor(target))
            else:
                event_system.loop.create_task(self.internal_delegate[event_name].parse_to_event(target).executor())

    def remove_delegate(self, target: Union[str, Event_T]):
        if isinstance(target, str):
            del self.internal_delegate[target]
        elif isinstance(target, TemplateEvent):
            del self.internal_delegate[target.__class__.__name__]
        else:
            del self.internal_delegate[target.__name__]



import traceback
from typing import Tuple, Dict, Type, Any, Union, Callable
from .entities.subscriber import Subscriber
from .entities.event import Insertable, ParamRet
from .exceptions import UndefinedRequirement, UnexpectedArgument, MultipleInserter, RepeatedInserter
from .utils import argument_analysis, run_always_await
from .entities.decorator import TemplateDecorator


async def await_exec_target(
        target: Union[Subscriber, Callable],
        event_data_handler: Union[Callable[[], ParamRet], ParamRet]
):
    is_subscriber = False
    if isinstance(target, Subscriber):
        is_subscriber = True
    decorators = target.decorators if is_subscriber else []
    callable_target = target.callable_target if is_subscriber else target
    target_param = target.params if is_subscriber else argument_analysis(callable_target)
    try:
        event_args = before_parser(decorators, event_data_handler)
        arguments = param_parser(target_param, event_args)
        real_arguments = after_parser(decorators, arguments)
        result = await run_always_await(callable_target, **real_arguments)
    except (UnexpectedArgument, UndefinedRequirement):
        traceback.print_exc()
        raise
    except Exception as e:
        traceback.print_exc()
        raise e
    return result


def decorator_before_handler(decorator: TemplateDecorator, event_args):
    if "before_parser" not in decorator.__disable__:
        for k, v in event_args.copy().items():
            if not decorator.may_target_type or (decorator.may_target_type and type(v) is decorator.may_target_type):
                result = decorator.supply_wrapper(k, v)
                if result is None:
                    continue
                event_args.update(result)
    return event_args


def inserter_handler(inserters: Tuple[Union[Type[Insertable], Insertable]]) -> Dict[str, Any]:
    extra: Dict[str, Any] = {}
    used_inserter = []
    for inserter in inserters:
        if inserter in used_inserter:
            raise RepeatedInserter("a inserter cannot insert twice")
        (event_inserter, event_args) = inserter.get_params()
        if event_inserter:
            raise MultipleInserter("a inserter cannot used another inserter")
        extra.update(event_args)
        used_inserter.append(inserter)
    return extra


def before_parser(decorators, event_data_handler):
    (event_inserter, event_args) = event_data_handler()
    try:
        if event_inserter:
            event_args.update(inserter_handler(event_inserter))
    except (RepeatedInserter, MultipleInserter):
        traceback.print_exc()
    finally:
        if decorators:
            for decorator in decorators:
                event_args = decorator_before_handler(decorator, event_args)
        return event_args


def param_parser(params, event_args):
    """
    将调用函数提供的参数字典与事件提供的参数字典进行比对，并返回正确的参数字典

    Args:
        params: 调用的函数的参数字典
        event_args: 函数可返回的参数字典
    Returns:
        函数需要的参数字典
    """
    arguments_dict = {}
    for name, annotation, default in params:
        if not annotation:
            raise UndefinedRequirement(f"a argument: {{<{annotation}> {name}: {default}}} without annotation")
        elif isinstance(default, str) and default in event_args:
            arguments_dict.setdefault(name, event_args[default])
        elif annotation.__name__ in event_args:
            arguments_dict.setdefault(name, event_args.get(annotation.__name__))
        elif name in event_args:
            arguments_dict.setdefault(name, event_args[name])
        elif default is not None:
            arguments_dict[name] = default
            if isinstance(default, Callable):
                arguments_dict[name] = default()
        if name not in arguments_dict:
            raise UnexpectedArgument(f"a unexpected extra argument: {{{annotation} {name}: {default}}}")

    return arguments_dict


def decorator_after_handler(decorator: TemplateDecorator, argument):
    if "after_parser" not in decorator.__disable__:
        for k, v in argument.copy().items():
            result = decorator.supply_wrapper(k, v)
            if result is None:
                continue
            argument.update(result)
    return argument


def after_parser(decorators, argument):
    if decorators:
        for decorator in decorators:
            argument = decorator_after_handler(decorator, argument)
    return argument



from abc import ABC, abstractmethod
from typing import overload, Callable


class TemplateCondition(ABC):
    """
    条件基类, 用于事件发生时判断是否响应
    """
    judge: Callable[[], bool]

    @staticmethod
    @overload
    @abstractmethod
    def judge(*args, **kwargs) -> bool:
        ...

    @classmethod
    @overload
    @abstractmethod
    def judge(cls, *args, **kwargs) -> bool:
        ...

    @abstractmethod
    def judge(self, *args, **kwargs) -> bool:
        """外部方法,用于判断该条件是否成立并返回结果; 该方法可以是 `stasticmethod`,`classmethod` 亦或是普通的方法/函数.
        唯一的要求是返回值必须为一布尔值
        """
        pass

    def __eq__(self, other: "TemplateCondition"):
        return self.__annotations__ == other.__annotations__




class UnexpectedArgument(Exception):
    pass


class UndefinedRequirement(Exception):
    pass


class MultipleInserter(Exception):
    pass


class RepeatedInserter(Exception):
    pass


class PropagationCancelled(Exception):
    pass



import asyncio
from typing import List, Union, Dict, Any
from .subscriber import Subscriber
from ..exceptions import PropagationCancelled
from ..handler import await_exec_target
from ..utils import Event_T


class EventDelegate:
    """
    订阅器相关的处理器
    """
    subscribers: List[Subscriber]
    bind_event: Event_T

    def __init__(self, event: Event_T):
        self.subscribers = []
        self.bind_event = event

    def __iadd__(self, other):
        if isinstance(other, Subscriber):
            if other.name in self.subscribers_names():
                raise ValueError(f"Function \"{other.name}\" for event: "
                                 f"{self.bind_event.__name__} has already registered!")
            self.subscribers.append(other)
            return self
        elif isinstance(other, List):
            for sub in other:
                self.__iadd__(sub)
            return self

    def __eq__(self, other: Union["EventDelegate", Event_T]):
        if isinstance(other, EventDelegate):
            return all([other.subscribers == self.subscribers, other.bind_event is self.bind_event])
        else:
            return other is self.bind_event

    def subscribers_names(self):
        return [i.name for i in self.subscribers]

    def parse_to_event(self, data: Union[Event_T, Dict[str, Any]]):
        if isinstance(data, dict):
            setattr(self, "wait_event", self.bind_event(**{k: v for k, v in data.items() if k != "type"}))
            return self
        setattr(self, "wait_event", data)
        return self

    async def executor(self, event: Event_T = None):
        current_event = getattr(self, "wait_event", event)
        event_params = current_event.get_params
        # self.subscribers.sort(key=lambda x: x.priority)
        coroutine = [
            await_exec_target(
                target,
                event_params
            )
            for target in self.subscribers
        ]
        (results, _) = await asyncio.wait(coroutine)
        for task in results:
            if task.exception().__class__ is PropagationCancelled:
                break



import setuptools

with open("README.rst", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="arclet-letoderea",
    version="0.2.2",
    author="ArcletProject",
    author_email="rf_tar_railt@qq.com",
    description="A high-performance, simple-structured event system, relies on asyncio",
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/rst",
    url="https://github.com/ArcletProject/Letoderea",
    packages=['arclet.letoderea', 'arclet.letoderea.entities'],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    keywords='asyncio, dict',
    python_requires='>=3.8'
)
